#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <termios.h>
#include <stdio.h>
#include <string.h>
#include <stdbool.h>
#include <stdlib.h>

#define SERIAL_BAUDRATE B115200
#define WAIT_1ST_SEC 5
#define WAIT_DEFAULT_SEC 1
#define AT_BUF_SIZE 185

static char _recv_buffer[AT_BUF_SIZE];
static int _recv_length = 0;
static struct termios tio_orig;
static bool has_tio_orig = false;

static void restore(int fd)
{
	if (has_tio_orig)
		tcsetattr(fd, TCSANOW, &tio_orig);

	has_tio_orig = false;
}

static bool setup(int fd, bool enable_flowctl)
{
	struct termios tio;
	int ret;

	ret = tcgetattr(fd, &tio_orig);
	if (ret) {
		perror("tcgetattr");
		return false;
	}
	has_tio_orig = true;

	memset (&tio, 0, sizeof(struct termios));

	if (enable_flowctl)
		tio.c_cflag &= ~(CBAUD | CSIZE | CSTOPB | PARENB | PARODD | CRTSCTS);
	else
		tio.c_cflag &= ~(CBAUD | CSIZE | CSTOPB | PARENB | PARODD);

	tio.c_iflag &= ~(IGNCR | ICRNL | IUCLC | INPCK | IXON | IXOFF | IXANY );
	tio.c_oflag &= ~(OPOST | OLCUC | OCRNL | ONLCR | ONLRET);
	tio.c_lflag &= ~(ICANON | ECHO | ECHOE | ECHONL);
	tio.c_cc[VMIN] = 1;
	tio.c_cc[VTIME] = 0;
	tio.c_cc[VEOF] = 1;
	tio.c_iflag |= IGNPAR;
	tio.c_cflag |= (CS8 | CLOCAL | CREAD | CRTSCTS);

	ret = cfsetispeed(&tio, SERIAL_BAUDRATE);
	if (ret < 0) {
		perror("cfsetispeed");
		return false;
	}

	ret = cfsetospeed(&tio, SERIAL_BAUDRATE);
	if (ret < 0) {
		perror("cfsetospeed");
		return false;
	}

	ret = tcflush(fd, TCIFLUSH);
	if (ret < 0)
		perror("tcflush");

	ret = tcsetattr(fd, TCSANOW, &tio);
	if (ret)
		perror("tcsetattr");

	return true;
}

static void close_serial(int fd);
static int open_serial(const char *port_name)
{
	int fd;

	fd = open(port_name, O_RDWR);
	if (fd == -1) {
		perror("open serial");
		return -1;
	}

	if (!setup(fd, true)) {
		close_serial(fd);
		return -1;
	}

	return fd;
}

static void close_serial(int fd)
{
	restore(fd);
	close(fd);
}

static int copy_valid_command(char *buf, char *end, int cut)
{
	int valid_length;

	valid_length = end - _recv_buffer;
	memcpy(buf, _recv_buffer, valid_length);
	valid_length += cut;

	if (valid_length == _recv_length) {
		memset(_recv_buffer, 0, AT_BUF_SIZE);
		_recv_length = 0;
	} else {
		memmove(_recv_buffer, &(_recv_buffer[valid_length]),
			AT_BUF_SIZE - valid_length);
		_recv_length -= valid_length;
		memset(&(_recv_buffer[AT_BUF_SIZE - valid_length]),
		       0, valid_length);
	}
	return valid_length;
}

static void cutCRandLF(char *p)
{
	if (!p)
		return;

	while(1) {
		if ((p[strlen(p) -1] == '\r') || (p[strlen(p) -1] == '\n'))
			p[strlen(p) -1] = '\0';
		else
			break;
	}
}

static int copy_at_command(char *buf)
{
	char *p = NULL;
	int cut = 0;
	int result;

	p = strstr(_recv_buffer, "\r\n");
	if (p)
		cut = 2;
	else {
		p = strchr(_recv_buffer, '\r');
		if (p)
			cut = 1;
		else {
			p = strchr(_recv_buffer, '\n');
			if (p)
				cut = 1;
			else
				return -1;
		}
	}

	result = copy_valid_command(buf, p, cut);
	if (result > 0)
		cutCRandLF(buf);

	return strlen(buf);
}

static bool has_received_command()
{
	return (_recv_length > 0);
}

int wait_data(int fd, int sec)
{
	struct timeval tv;
	fd_set readfds;
	int ret;

	FD_ZERO(&readfds);
	FD_SET(fd, &readfds);

	if (sec == 0) {
		tv.tv_sec = 0;
		tv.tv_usec = 1;
	} else {
		tv.tv_sec = sec;
		tv.tv_usec = 0;
	}

	ret = select(fd + 1, &readfds, NULL, NULL, &tv);

	if (ret == -1) {
		perror("select");
		return -1;
	}

	if (!ret)
		return -1;

	if (FD_ISSET(fd, &readfds))
		return 0;

	return -1;
}

static int read_line(int fd, char *buf, int timeout_sec)
{
	int current_length = 0;
	int sec = timeout_sec;
	int result;

	if ((_recv_length > 0) &&
	    (strchr(_recv_buffer, '\r') || strchr(_recv_buffer, '\n'))) {
		result = copy_at_command(buf);
		if (result == 0)
			result = copy_at_command(buf);
		return result;
	}

	while (wait_data(fd, sec) == 0) {
		current_length = read(fd, &(_recv_buffer[_recv_length]),
				      AT_BUF_SIZE - 1 - _recv_length);
		if (current_length < 0) {
			perror("read");
			return -1;
		}

		if (current_length == 0)
			return -1;

		_recv_length += current_length;

		if (strchr(_recv_buffer, '\r') || strchr(_recv_buffer, '\n')) {
			result = copy_at_command(buf);
			if (result == 0)
				result = copy_at_command(buf);
			return result;
		}

		sec = WAIT_DEFAULT_SEC;
	}
	return -1;
}

static int atcmd_send(const char *port_name, const char *command, bool echo)
{
	int read_size;
	int result = -1;
	char buf[AT_BUF_SIZE];
	int sec = WAIT_1ST_SEC;
	int fd;

	fd = open_serial(port_name);
	if (fd < 0)
		return -1;

	/* dummy read */
	memset(buf, 0, AT_BUF_SIZE);
	do {
		read_size = read_line(fd, buf, 0);
	} while (read_size > 0);

	/* send command */
	snprintf(buf, AT_BUF_SIZE, "%s\r\n", command);
	write(fd, buf, strlen(command)+2);

	/* read response */
	while (1) {
		if (!has_received_command() && (wait_data(fd, sec) != 0)) {
			result = -1;
			break;
		}

		sec = WAIT_DEFAULT_SEC;
		memset(buf, 0, AT_BUF_SIZE);
		read_size = read_line(fd, buf, sec);
		if (read_size <= 0) {
			fprintf(stderr, "error: timed out.\n");
			result = -1;
			break;
		}

		sec = WAIT_DEFAULT_SEC;
		if (strcmp(buf, command) == 0) /* command echo */
			continue;
		else if (strstr(buf, "OK")) {
			result = 0;
			fprintf(stderr, "OK\n");
			break;
		} else if (strstr(buf, "ERROR")) {
			result = -1;
			fprintf(stderr, "ERROR\n");
			break;
		} else if (echo)
			fprintf(stderr, "%s\n", buf);
	}

	close_serial(fd);
	return result;
}

int main(int argc, char *argv[])
{
	int result;

	if (argc == 3)
		result = atcmd_send(argv[1], argv[2], false);
	else if (argc == 4 && strcmp("echo", argv[3]) == 0)
		result = atcmd_send(argv[1], argv[2], true);
	else {
		fprintf(stderr, "send-at: port command [echo]\n");
		exit(1);
	}

	exit(result?1:0);
}
